# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .job import SinkJob
from .noderun import NodeRun
from .. import exceptions
from ..core.samples import SampleItem, SampleValue, SampleState
from ..datatypes import types
from ..helpers import log

__all__ = ['SinkNodeRun']


class SinkNodeRun(NodeRun):
    """
    Class which handles where the output goes. This can be any kind of file, e.g.
    image files, textfiles, config files, etc.
    """

    __dataschemafile__ = 'SinkNodeRun.schema.json'
    _JobType = SinkJob

    def __init__(self, node, parent):
        """ Instantiation of the SinkNodeRun.

        :param fastr.planning.node.Node node: The Node that this Run is based on.
        :param NetworkRun parent: The NetworkRun that this NodeRun belongs to
        :return: newly created sink node run
        """
        NodeRun.__init__(self, node, parent)

        # Set the DataType
        self.datatype = node.datatype
        self.url = node.url

    def __getstate__(self):
        state = super(SinkNodeRun, self).__getstate__()
        state['url'] = self.url
        return state

    def __setstate__(self, state):
        super(SinkNodeRun, self).__setstate__(state)
        self.url = state['url']

    @property
    def datatype(self):
        """
        The datatype of the data this sink can store.
        """
        return self.inputs['input'].datatype

    @datatype.setter
    def datatype(self, value):
        """
        The datatype of the data this sink can store (setter).
        """
        self.inputs['input'].datatype = value

    @property
    def input(self):
        """
        The default input of the sink NodeRun
        """
        return self.inputs['input']

    @input.setter
    def input(self, value):
        """
        The default input of the sink NodeRun (setter)
        """
        self.inputs['input'] = value

    def _execute(self):
        """
        Execute the sink node and create the jobs that need to run

        :return: list of jobs to run
        :rtype: list of :py:class:`Jobs <fastr.execution.job.Job>`
        """
        self.update(False, False)

        for sample_item in self.inputs['input'].items():
            for cardinality_nr, value in enumerate(sample_item.data.sequence_part()):
                job = self.create_job(sample_item.id,
                                      sample_item.index,
                                      {
                                          'input': SampleItem(
                                              sample_item.index,
                                              sample_item.id,
                                              SampleValue({0: (value,)},),
                                              failed_annotations=sample_item.failed_annotations,
                                              status=sample_item.status,
                                          ),
                                          'cardinality': cardinality_nr
                                      }, sample_item.jobs, sample_item.status)
                yield job

        self.drained = True

    def set_data(self, data):
        """
        Set the targets of this sink node.

        :param data: the targets rules for where to write the data
        :type data: dict or list of urls

        The target rules can include a few fields that can be filled out:

        =========== ==================================================================
        field       description
        =========== ==================================================================
        sample_id   the sample id of the sample written in string form
        cardinality the cardinality of the sample written
        ext         the extension of the datatype of the written data, including the .
        extension   the extension of the datatype of the written data, excluding the .
        network     the id of the network the sink is part of
        node        the id of the node of the sink
        timestamp   the iso formatted datetime the network execution started
        uuid        the uuid of the network run (generated using uuid.uuid1)
        =========== ==================================================================

        An example of a valid target could be:

        .. code-block:: python

          >>> target = 'vfs://output_mnt/some/path/image_{sample_id}_{cardinality}{ext}'

        .. note::
            The ``{ext}`` and ``{extension}`` are very similar but are both offered.
            In many cases having a ``name.{extension}`` will feel like the correct way
            to do it. However, if you have DataTypes with and without extension that
            can both exported by the same sink, this would cause either ``name.ext`` or
            ``name.`` to be generated. In this particular case ``name{ext}`` can help
            as it will create either ``name.ext`` or ``name``.

        .. note::
            If a datatype has multiple extensions (e.g. .tiff and .tif) the first
            extension defined in the extension tuple of the datatype will be used.
        """
        if isinstance(data, str):
            try:
                data.format(sample_id='dummy',
                            cardinality=0,
                            ext='.ext',
                            extension='ext',
                            network='network',
                            node='node',
                            timestamp='timestamp',
                            uuid='uuid')
            except KeyError as error:
                raise exceptions.FastrValueError(
                    ('Using unknown substitution "{}" in SinkData "{}", valid'
                     ' substitution fields are: sample_id, cardinality,'
                     ' ext').format(error.message, data)
                )
            self.url = data
        else:
            raise exceptions.FastrTypeError(
                'Invalid datatype for SinkNode data, expected str but got {}!'.format(type(data).__name__)
            )

    def set_result(self, job, failed_annotation):
        """
        Incorporate result of a sink job into the Network.

        :param Type job: job of which the result to store
        :param set failed_annotation: A set of annotations, None if no errors else containing
                                      a tuple describing the errors
        """
        super(SinkNodeRun, self).set_result(job, failed_annotation)

        if self.id not in self.parent.sink_results:
            self.parent.sink_results[self.id] = {}

        # First check if Sample is MISSING, if not check the Sink for error,
        # if no error then use the input sample status
        input_sample = self.inputs['input'][job.sample_index]
        if input_sample.status == SampleState.MISSING:
            status = SampleState.MISSING
        elif job.status.error:
            status = SampleState.FAILED
        else:
            status = input_sample.status

        self.parent.sink_results[self.id][job.sample_id] = (job, status, failed_annotation)

    def create_job(self, sample_id, sample_index, job_data, job_dependencies, status, **kwargs):
        """
        Create a job for a sink based on the sample id, job data and job dependencies.

        :param sample_id: the id of the corresponding sample
        :type sample_id: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :param dict job_data: dictionary containing all input data for the job
        :param job_dependencies: other jobs that need to finish before this job can run
        :return: the created job
        :rtype: :py:class:`Job <fastr.execution.job.Job>`
        """

        substitutions = {'sample_id': sample_id,
                         'cardinality': job_data['cardinality'],
                         'timestamp': self.parent.timestamp.isoformat(),
                         'uuid': self.parent.uuid,
                         'network': self.parent.id,
                         'node': self.id}

        job = super(SinkNodeRun, self).create_job(sample_id, sample_index, job_data, job_dependencies, status,
                                                  substitutions=substitutions, **kwargs)

        self.jobs[sample_id] = job
        return job

    def _wrap_arguments(self, job_data, sample_id, sample_index):
        """
        Wrap arguments into a list of tuples that the execution script can parse

        :param dict job_data: dictionary containing all input data for the job
        :param sample_id: the id of the corresponding sample
        :type sample_id: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the wrapped arguments in a tuple with the form ``(inputs, outputs)``

        .. note::
            For a SinkNodeRun this function adds a few default (hidden) arguments
        """
        arguments = super(SinkNodeRun, self)._wrap_arguments(job_data, sample_id, sample_index)
        arguments[0]['output'] = SampleItem(sample_index, sample_id, types['String'](self.url))
        arguments[0]['datatype'] = SampleItem(sample_index, sample_id, types['String'](self.datatype.id))

        log.debug('Wrapped Sink arguments to {}'.format(arguments))

        return arguments



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from logging import LogRecord

from fastr.abc.baseplugin import Plugin
from fastr.execution.job import Job
from fastr.execution.networkrun import NetworkRun
from fastr.helpers.events import EventType, register_listener, remove_listener


class ReportingPlugin(Plugin):
    """
    Base class for all reporting plugins. The plugin has a number of methods that
    can be implemented that will be called on certain events. On these events the
    plugin can inspect the presented data and take reporting actions.
    """
    _instantiate = True

    def activate(self):
        if self.job_updated.__qualname__ != 'ReportingPlugin.job_updated':
            register_listener(EventType.job_updated, self.job_updated)

        if self.run_started.__qualname__ != 'ReportingPlugin.run_started':
            register_listener(EventType.run_started, self.run_started)

        if self.run_finished.__qualname__ != 'ReportingPlugin.run_finished':
            register_listener(EventType.run_finished, self.run_finished)

        if self.log_record_emitted.__qualname__ != 'ReportingPlugin.log_record_emitted':
            register_listener(EventType.log_record_emitted, self.log_record_emitted)

    def deactivate(self):
        remove_listener(EventType.job_updated, self.job_updated)
        remove_listener(EventType.run_started, self.run_started)
        remove_listener(EventType.run_finished, self.run_finished)
        remove_listener(EventType.log_record_emitted, self.log_record_emitted)

    def job_updated(self, job: Job):
        pass

    def run_started(self, run: NetworkRun):
        pass

    def run_finished(self, run: NetworkRun):
        pass

    def log_record_emitted(self, record: LogRecord):
        pass





# Copyright 2011-2017 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some helper functions that aid with NFS file sync issues.
"""
from . import config, log
import time
import hashlib
import os.path
from glob import glob


def filesynchelper_enabled():
    return len(config.filesynchelper_url) > 0


if filesynchelper_enabled():
    from redis import Redis


class FileSyncHelper():
    _namespace = 'filesynchelper'
    _redis = Redis.from_url(config.filesynchelper_url) if filesynchelper_enabled() else None

    def __init__(self):
        pass

    def job_finished(self, jobfile):
        key = self._generate_key_for_string('joblock' + str(jobfile))
        self._redis.setex(key, '', 300)

    def wait_for_job(self, jobfile):
        key = self._generate_key_for_string('joblock' + str(jobfile))
        # wait for job or timeout
        timeoutafter = time.time() + 300
        exists = self._redis.exists(key)
        while not exists or time.time() > timeoutafter:
            time.sleep(5)  # wake-up every 5 seconds
            exists = self._redis.exists(key)
        self._redis.delete(key)

    def wait_for_pickle(self, url, timeout=300):
        log.debug('waiting for pickle {}'.format(url))
        # wait for file or timeout
        timeoutafter = time.time() + timeout
        exists = self.has_file_promise(url)
        while not exists or time.time() > timeoutafter:
            time.sleep(5)  # wake-up every 5 seconds
            exists = self.has_file_promise(url)

        return exists

    def store(self, url, data):
        key = self._generate_key_for_string(str(url))
        log.debug('storing {} -> {}'.format(url, key))
        self._redis.delete(key)
        return self._redis.setex(
            key,
            data,
            300
        )

    def load(self, url):
        key = self._generate_key_for_string(str(url))
        log.debug('loading {} from {}'.format(url, key))
        return self._redis.get(key)

    def _generate_key_for_string(self, input_string):
        return self._namespace + self._generate_hash_from_string(input_string)

    def _generate_hash_from_string(self, input_string):
        return hashlib.sha256(input_string).hexdigest()

    def make_file_promise(self, url):
        # ttl of 86400 seconds is 1 day
        # we don't really care about the value if suburls == None, thus an empty string is stored
        key = self._generate_key_for_string(url)
        log.debug('making file promise {} -> {}'.format(url, key))
        if os.path.isdir(url):
            dir = url
        elif url.startswith('vfs://') and os.path.isdir(fastr.vfs.url_to_path(url)):
            dir = fastr.vfs.url_to_path(url)
        else:
            dir = None

        if dir is not None:
            val = ','.join([self._generate_hash_from_string(str(suburl)) for suburl in self._glob_dir(dir)])
        else:
            val = ''

        self._redis.setex(
            key,
            val,
            300
        )

    def has_file_promise(self, url):
        # check if key exists, if it does we have a file promise
        result = self._redis.exists(self._generate_key_for_string(str(url)))
        log.debug('has_file_promise {} {} {}'.format(url, result, self._generate_key_for_string(str(url))))
        return result

    def wait_for_vfs_url(self, vfs_url, timeout=300):
        log.debug('wait_for_vfs_url {}'.format(vfs_url))
        suburls = self._get_suburl_hashes(vfs_url)
        return self._wait_for_file_and_suburls(fastr.vfs.url_to_path(vfs_url), suburls, timeout)

    def wait_for_file(self, path, timeout=300):
        log.debug('wait_for_file {}'.format(path))
        suburls = self._get_suburl_hashes(path)
        return self._wait_for_file_and_suburls(path, suburls, timeout)

    def _get_suburl_hashes(self, path):
        if self.has_file_promise(path):
            suburls = self._redis.get(self._generate_key_for_string(str(path)))
            if len(suburls) > 0:
                return suburls.split(',')
        return None

    def _glob_dir(self, dir):
        if not dir.endswith('/'):
            dir = dir + '/'
        return [y[len(dir):] for x in os.walk(dir) for y in glob(os.path.join(x[0], '*'))]

    def _wait_for_file_and_suburls(self, path, suburls, timeout):
        log.debug('waiting for {}'.format(path))
        # wait for file or timeout
        timeoutafter = time.time() + timeout
        fileexists = os.path.isfile(path)
        while not fileexists or time.time() > timeoutafter:
            time.sleep(5)  # wake-up every 5 seconds
            fileexists = (os.path.isdir(path) if suburls is not None else os.path.isfile(path))

        if suburls is not None:
            fileexists = False
            lookup = {}
            while len(suburls) > 0 or time.time() > timeoutafter:
                files = self._glob_dir(path)
                for file in files:
                    if file not in lookup:
                        lookup[file] = self._generate_hash_from_string(str(file))
                    if lookup[file] in suburls:
                        suburls.remove(lookup[file])
                time.sleep(5) # wake-up every 5 seconds

            if len(suburls) == 0:
                fileexists = True

        return fileexists



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import tempfile

import pytest

import fastr
from fastr.abc.serializable import load, save
from fastr.core.tool import Tool
from fastr.core.version import Version


@pytest.fixture(scope="module")
def tool_addint():
    return fastr.tools['fastr/math/Add:1.0', '1.0']


@pytest.fixture(scope="module")
def tool_sum():
    return fastr.tools['fastr/math/Sum:1.0', '1.0']


def test_tool_eq_operator(tool_addint, tool_sum):
    # Create copies for testing equal operator
    tool_addint_copy = copy.deepcopy(tool_addint)
    tool_sum_copy = copy.deepcopy(tool_sum)

    assert tool_addint == tool_addint
    assert tool_addint == tool_addint_copy
    assert tool_sum == tool_sum_copy
    assert tool_addint != tool_sum
    assert tool_addint_copy != tool_sum


def test_tool_getstate_setstate(tool_addint, tool_sum):
    #  Dump state and rebuild in a clean object
    addint_state = tool_addint.__getstate__()
    addint_rebuilt = Tool.__new__(Tool)
    addint_rebuilt.__setstate__(addint_state)
    assert tool_addint == addint_rebuilt

    #  Dump state and rebuild in a clean object
    sum_state = tool_sum.__getstate__()
    sum_rebuilt = Tool.__new__(Tool)
    sum_rebuilt.__setstate__(sum_state)
    assert tool_sum == sum_rebuilt


def test_tool_serializing(tool_addint, tool_sum):
    addint_json = tool_addint.serialize()
    addint_rebuilt = Tool.deserialize(addint_json)
    assert tool_addint == addint_rebuilt

    sum_json = tool_sum.serialize()
    sum_rebuilt = Tool.deserialize(sum_json)
    assert tool_sum == sum_rebuilt


def test_tool_save_load(tmp_path, tool_addint, tool_sum):
    addint_path = tmp_path / 'addint.yaml'
    save(tool_addint, addint_path)
    addint_rebuilt = load(addint_path)
    assert tool_addint == addint_rebuilt

    sum_path = tmp_path / 'sum.yaml'
    save(tool_sum, sum_path)
    sum_rebuilt = load(sum_path)
    assert tool_sum == sum_rebuilt


def test_toolversions():
    toolname = 'fastr/Source:1.0'
    version = Version("1.0")
    tool = fastr.tools[toolname, version]
    assert fastr.tools.toolversions(tool)[0] == version
    assert fastr.tools.toolversions(toolname)[0] == version
    assert len(fastr.tools.toolversions(toolname)) == 1  # "Expected number of versions for tool is 1

    # Maybe a unique id should be used here, but I think it very unlikely that this toolname will be encountered.
    assert fastr.tools.toolversions('097G)(A&*BD)(A7vA)(F67vAD)F967vAF-67vAD)F(86vAFD096vAF)97v') == []
    assert isinstance(fastr.tools.toolversions(tool), list)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from fastr import version
from fastr.helpers import log
from fastr.execution.networkrun import NetworkRun
from fastr.plugins.reportingplugin import ReportingPlugin


class SimpleReport(ReportingPlugin):
    def run_finished(self, run: NetworkRun):
        log.info('===== RESULTS =====')
        result = True
        for sink_node, sink_data in sorted(run.sink_results.items()):
            nr_failed = sum(x[1].name == 'FAILED' for x in sink_data.values())
            nr_missing = sum(x[1].name == 'MISSING' for x in sink_data.values())
            nr_success = sum(x[1].name == 'VALID' for x in sink_data.values())

            if nr_failed > 0:
                result = False

            log.info(f'{sink_node}: {nr_success} success / {nr_missing} missing / {nr_failed} failed')
        log.info('===================')

        if not result:
            sink_result_file = os.path.join(run.tmpdir, run.SINK_DUMP_FILE_NAME)

            log.warning(
                """There were failed samples in the run, to start debugging you can run:

    fastr trace {sink_data_file} --sinks

see the debug section in the manual at https://fastr.readthedocs.io/en/{branch}/static/user_manual.html#debugging for more information.""".format(
                    sink_data_file=sink_result_file,
                    branch='default' if version.git_branch == 'default' else 'develop',
                )
            )



import argparse
import os
import zipfile

from fastr.utils.cmd import add_parser_doc_link


def get_parser():
    parser = argparse.ArgumentParser(
        description="Create a dump of a network run directory that contains the"
                    " most important information for debugging. This includes a"
                    " serialization of the network, all the job command and"
                    " result files, the extra job information files and the "
                    " provenance files. No data files will be included, but note"
                    " that if jobs get sensitive information passed via the "
                    " command line this will be included in the job files.")
    parser.add_argument('indir', metavar='RUNDIR',
                        help='The run directory to dump')
    parser.add_argument('outfile', metavar='DUMP.zip',
                        help='The file to place the dump in')
    return parser


def create_zip(directory, output_file):
    with zipfile.ZipFile(output_file, 'w', compression=zipfile.ZIP_DEFLATED) as zip_file:
        zip_file.write(os.path.join(directory, './__sink_data__.json'), './__sink_data__.json')
        for path, _, files in os.walk(directory):
            for filename in files:
                if filename.startswith('__fastr'):
                    file_path = os.path.join(path, filename)
                    relative_path = file_path.replace(directory, './', 1)
                    zip_file.write(file_path, relative_path)


def main():
    """
    Dump the contents of a network run tempdir into a zip for remote assistance
    """
    parser = add_parser_doc_link(get_parser(), __file__)
    args = parser.parse_args()

    create_zip(args.indir, args.outfile)


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import imp
import inspect
import os
import sys
import traceback
from typing import Type, TypeVar, Generic
from abc import abstractmethod

from .basemanager import BaseManager
from .baseplugin import BasePlugin, PluginState
from .. import exceptions
from ..helpers import log


plugin_option_type = collections.namedtuple('plugin_option_type', ['filename', 'name', 'namespace', 'id'])

PluginType = TypeVar('PluginType')


class BasePluginManager(BaseManager[str, PluginType], Generic[PluginType]):
    """
    Baseclass for PluginManagers, need to override the self._plugin_class
    """

    def __init__(self, path=None, recursive=False, module=None):
        """
        Create a BasePluginManager and scan the give path for matching plugins

        :param str path: path to scan
        :param bool recursive: flag to indicate a recursive search
        :param module module: the module to register plugins into
        :return: newly created plugin manager
        :raises FastrTypeError: if self._plugin_class is set to a class not
                                 subclassing BasePlugin
        """
        self._loaded_plugins = {}
        self._plugin_options = {}
        self._module = module

        super(BasePluginManager, self).__init__(path, recursive)

        if not issubclass(self.plugin_class, BasePlugin):
            raise exceptions.FastrTypeError(
                'Plugin type to manage ({}) not a valid plugin! (needs to be subclass of BasePlugin)'.format(
                    self.plugin_class.__name__
                )
            )

    def set_module(self, module):
        if self._module is not None:
            raise ValueError('Cannot reset module!')
        self._module = module

    def test_plugin(self, plugin):
        # Since we cannot know what Plugins might throw, catch all
        # pylint: disable=broad-except
        try:
            # Let the Plugin think it is loaded, or it will refuse to instantiate
            plugin.set_status(PluginState.preload, 'Set to PreLoad to perform testing')
            plugin.test()
            # Let the Plugin think it is loaded, or it will refuse to instantiate
            plugin.set_status(PluginState.loaded, 'Testing successful, loaded properly')
        # Register the configuration for the plugin
        except Exception as exception:
            log.info('Could not load plugin file {}\n{}'.format(plugin.filename, exception))
            exc_type, _, _ = sys.exc_info()
            exc_info = traceback.format_exc()
            log.debug('Encountered exception ({}) during instantiation of the plugin:\n{}'.format(
                exc_type.__name__, exc_info)
            )
            exception_stacktrace = ('Encountered exception ({}) during'
                                    ' instantiation of the plugin:\n{}').format(exc_type.__name__, exc_info)
            exception_message = '[{}] {}'.format(plugin.fullid, exception)
            plugin.set_status(PluginState.failed, exception_message, exception_stacktrace)

    def __getitem__(self, key: str) -> PluginType:
        """
        Retrieve item from BaseManager

        :param key: the key of the item to retrieve
        :return: the value indicated by the key
        :raises FastrKeyError: if the key is not found in the BaseManager
        """
        try:
            plugin = super(BasePluginManager, self).__getitem__(key)
        except exceptions.FastrKeyError:
            self.load_plugin(key.lower())

            plugin = super(BasePluginManager, self).__getitem__(key)

        if plugin.status not in [PluginState.loaded, PluginState.failed]:
            self.test_plugin(plugin)

        return plugin

    @property
    @abstractmethod
    def plugin_class(self) -> Type[PluginType]:
        """
        The class from which the plugins must be subclassed
        """
        raise exceptions.FastrNotImplementedError

    @property
    def _item_extension(self) -> str:
        """
        Plugins should be loaded from files with a .py extension
        """
        return '.py'

    @property
    def _instantiate(self) -> bool:
        """
        Flag indicating that the plugin should be instantiated prior to saving
        """
        return True

    def _print_key(self, key):
        print_out = (self[key].status.value, key)

        return print_out

    def _print_value(self, val: PluginType) -> str:
        """
        Function for printing values (plugins) in this manager

        :param BasePlugin val: value to print
        :return: print representation
        :rtype: str
        """
        if val._instantiate:
            val = type(val)

        print_out = '<{}: {}>'.format(val.__bases__[0].__name__, val.__name__)
        return print_out

    def _load_item(self, filepath: str, namespace: str):
        """
        Load a plugin

        :param str filepath: path of the plugin to load
        """
        name = os.path.basename(filepath)
        name = os.path.splitext(name)[0].lower()

        value = plugin_option_type(filename=filepath, name=name, namespace=namespace, id=None)

        self._plugin_options[name] = value

    def populate(self):
        """
        Populate the manager with the data. This is a method that will be
        called when the Managers data is first accessed. This way we avoid
        doing expensive directory scans when the data is never requested.
        """
        super().populate()

        # Use list to avoid changing dictionary during loop
        for plugin_key in list(self._plugin_options.keys()):
            if plugin_key in self._plugin_options:
                self.load_plugin(plugin_key)

    def load_plugin(self, plugin_key: str):
        plugin_option = self._plugin_options[plugin_key]
        filepath = plugin_option.filename

        # Since we cannot know what Plugins might throw, catch all
        # pylint: disable=broad-except
        try:
            filebase, _ = os.path.splitext(os.path.basename(filepath))
            temp_module = imp.load_source(filebase, filepath)
            for name, obj in inspect.getmembers(temp_module):
                if not inspect.isclass(obj):
                    continue

                if filebase.lower() != obj.__name__.lower():
                    log.debug('Plugin name and module do not match ({} vs {})'.format(obj.__name__, filebase))
                    continue

                if not issubclass(obj, self.plugin_class):
                    log.debug('{} is not a subclass of {}'.format(obj, self.plugin_class))
                    continue

                obj.filename = filepath
                if inspect.isabstract(obj):
                    log.debug('Skipping abstract Plugin: {} ({})'.format(name, filepath))
                    continue

                if obj.status == PluginState.uninitialized:
                    obj.register_configuration()
                elif obj.status not in (PluginState.loaded, PluginState.failed):
                    log.warning('Invalid Plugin status: {}!'.format(obj.status))

                # Save the source in the obj
                obj.set_code(inspect.getsource(obj))
                obj.module = temp_module

                if obj.instantiate:
                    log.debug('Store instantiated plugin')
                    self.test_plugin(obj)
                    if obj.status == PluginState.loaded:
                        self._store_item(name, obj())
                    else:
                        self._store_item(name, obj)
                else:
                    log.debug('Store uninstantiated plugin')
                    self._store_item(name, obj)

        except Exception as exception:
            log.warning('Could not load {} file {}\n{}'.format(self.plugin_class.__name__, filepath, exception))
            exc_type, _, _ = sys.exc_info()
            exc_info = traceback.format_exc()
            log.info('Encountered exception ({}) during loading of the plugin:\n{}'.format(exc_type.__name__, exc_info))
        finally:
            del self._plugin_options[plugin_key]



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import queue
import sys
import threading
import time

from bidict import bidict
import fastr
from fastr import exceptions
from fastr.abc.baseplugin import PluginState

try:
    import drmaa
    load_drmaa = True
except (ImportError, RuntimeError, OSError):
    load_drmaa = False

from fastr.execution.job import JobState
from fastr.plugins.executionplugin import ExecutionPlugin
from fastr.helpers.classproperty import classproperty


class FastrDRMAANotFoundError(exceptions.FastrImportError):
    """
    Indicate the DRMAA module was not found on the system.
    """
    pass


class FastrDRMAANotFunctionalError(exceptions.FastrError):
    """
    Indicate DRMAA is found but creating a session did not work
    """
    pass


class DRMAAExecution(ExecutionPlugin):
    """
    A DRMAA execution plugin to execute Jobs on a Grid Engine cluster. It uses
    a configuration option for selecting the queue to submit to. It uses the
    python ``drmaa`` package.

    .. note::

        To use this plugin, make sure the ``drmaa`` package is installed and
        that the execution is started on an SGE submit host with DRMAA
        libraries installed.

    .. note::

        This plugin is at the moment tailored to SGE, but it should be fairly
        easy to make different subclasses for different DRMAA supporting
        systems.
    """
    if not load_drmaa:
        _status = (PluginState.failed, 'Could not load DRMAA module required for cluster communication')

    # DRMAA Supports cancelling jobs, job dependencies and hold release actions
    SUPPORTS_CANCEL = True
    SUPPORTS_DEPENDENCY = True
    SUPPORTS_HOLD_RELEASE = True
    CANCELS_DEPENDENCIES = False

    GE_NATIVE_SPEC = {
        'WD': '-wd {workdir}',
        'QUEUE': '-q {queue}',
        'WALLTIME': '-l h_rt={walltime}',
        'MEMORY': '-l h_vmem={memory}',
        'NCORES': '-pe smp {ncores:d}',
        'OUTPUTLOG': '-o {outputlog}',
        'ERRORLOG': '-e {errorlog}',
        'DEPENDS': '-hold_jid {hold_list}',
        'DEPENDS_SEP': ',',
        'HOLD': '-h',
    }

    TORQUE_NATIVE_SPEC = {
        'CWD': '',
        'QUEUE': '-q {queue}',
        'WALLTIME': '-l walltime={walltime}',
        'MEMORY': '-l mem={memory}',
        'NCORES': '-l procs={ncores:d}',
        'OUTPUTLOG': '-o {outputlog}',
        'ERRORLOG': '-e {errorlog}',
        'DEPENDS': '-W depend=afterok:{hold_list}',
        'DEPENDS_SEP': ':',
        'HOLD': '-h',
    }

    NATIVE_SPEC = {
        'grid_engine': GE_NATIVE_SPEC,
        'torque': TORQUE_NATIVE_SPEC,
    }

    def __init__(self, finished_callback=None, cancelled_callback=None):
        super(DRMAAExecution, self).__init__(finished_callback, cancelled_callback)

        # Some default
        self.default_queue = fastr.config.drmaa_queue
        self.max_jobs = fastr.config.drmaa_max_jobs
        self.engine = fastr.config.drmaa_engine
        self.regression_check_interval = fastr.config.drmaa_job_check_interval
        self.num_undetermined_to_fail = fastr.config.drmaa_num_undetermined_to_fail

        # Create the DRMAA session
        try:
            self.session = drmaa.Session()
            self.session.initialize()
        except drmaa.errors.DrmaaException as exception:
            raise FastrDRMAANotFunctionalError('Encountered an error when creating DRMAA session: [{}] {}'.format(
                exception.__class__,
                str(exception)
            ))

        fastr.log.debug('A DRMAA session was started successfully')
        response = self.session.contact
        fastr.log.debug('session contact returns: ' + response)

        # Create job translation table
        self.job_lookup_fastr_drmaa = bidict()
        self.job_lookup_lock = threading.Lock()

        # Track jobs of which we cannot determine state and count many times
        # this happened. If only a few times, it might be communications
        # hiccups, but if consistent, the jobs probably died or something.
        self.undetermined_jobs = {}

        # Create even queue lock
        self.submit_queue = queue.Queue()
        self.finished_queue = queue.Queue()

        # Make sure we can regularly check if the threads are still alive
        self.last_thread_check = time.time()
        self.thread_check_lock = threading.Lock()

        # Threads for plugin
        self.collector_thread = None
        self.callback_thread = None
        self.submitter_thread = None
        self.checker_thread = None
        self.ensure_threads()

    def check_threads(self):
        """
        Check if the threads are still alive, but make sure it is only done once per minute
        """
        with self.thread_check_lock:
            # Check only once per minute
            if time.time() - self.last_thread_check < 60:
                return

            self.last_thread_check = time.time()
            self.ensure_threads()

    def ensure_threads(self):
        """
        Start thread if not defined, or restart if they somehow died accidentallyy
        """
        # If plugin is not running, no need for threads
        if not self.running:
            return

        if self.collector_thread is None or not self.collector_thread.is_alive():
            fastr.log.debug('Creating job collector')
            self.collector_thread = threading.Thread(name='DRMAAJobCollector-0', target=self.collect_jobs, args=())
            self.collector_thread.daemon = True
            fastr.log.debug('Starting job collector')
            self.collector_thread.start()

        if self.callback_thread is None or not self.callback_thread.is_alive():
            fastr.log.debug('Creating job callback processor')
            self.callback_thread = threading.Thread(name='DRMAAJobCallback-0', target=self.dispatch_callbacks, args=())
            self.callback_thread.daemon = True
            fastr.log.debug('Starting job callback processor')
            self.callback_thread.start()

        if self.submitter_thread is None or not self.submitter_thread.is_alive():
            fastr.log.debug('Creating job submitter')
            self.submitter_thread = threading.Thread(name='DRMAAJobSubmitter-0', target=self.submit_jobs, args=())
            self.submitter_thread.daemon = True
            fastr.log.debug('Starting job submitter')
            self.submitter_thread.start()

        if self.checker_thread is None or not self.checker_thread.is_alive():
            fastr.log.debug('Creating job regression checker')
            self.checker_thread = threading.Thread(name='DRMAAJobChecker-0', target=self.regression_check, args=())
            self.checker_thread.daemon = True
            fastr.log.debug('Starting job regression checker')
            self.checker_thread.start()

    @classproperty
    def configuration_fields(cls):
        return {
            "drmaa_queue": (str, "week", "The default queue to use for jobs send to the scheduler"),
            "drmaa_max_jobs": (int, 0, "The maximum jobs that can be send to the scheduler"
                                       " at the same time (0 for no limit)"),
            "drmaa_engine": (str, "grid_engine", "The engine to use (options: grid_engine, torque"),
            "drmaa_job_check_interval": (int, 900, "The interval in which the job checker will start"
                                                   " to check for stale jobs"),
            "drmaa_num_undetermined_to_fail": (int, 3, "Number of consecutive times a job state has be"
                                                       " undetermined to be considered to have failed")
        }

    @classmethod
    def test(cls):
        if not load_drmaa:
            raise FastrDRMAANotFoundError('Could not import the required drmaa for this plugin')

    @property
    def spec_fields(self):
        return self.NATIVE_SPEC[self.engine]

    @property
    def n_current_jobs(self):
        return len(self.job_lookup_fastr_drmaa)

    def cleanup(self):
        # Stop submissions and callbacks
        super(DRMAAExecution, self).cleanup()

        # See if there are leftovers in the job translation table that can be cancelled
        while self.n_current_jobs > 0:
            with self.job_lookup_lock:
                fastr_job_id, drmaa_job_id = self.job_lookup_fastr_drmaa.popitem()

            fastr.log.info('Terminating left-over job {}'.format(drmaa_job_id))
            self._terminate_job(drmaa_job_id, retry=0)

        fastr.log.debug('Stopping DRMAA executor')
        # Destroy DRMAA
        try:
            self.session.exit()
            fastr.log.debug('Exiting DRMAA session')
        except drmaa.NoActiveSessionException:
            pass

        if self.collector_thread.is_alive():
            fastr.log.debug('Terminating job collector thread')
            self.collector_thread.join()
        if self.submitter_thread.is_alive():
            fastr.log.debug('Terminating job submitter thread')
            self.submitter_thread.join()
        if self.checker_thread.is_alive():
            fastr.log.debug('Terminating job checker thread')
            self.checker_thread.join()
        if self.callback_thread.is_alive():
            fastr.log.debug('Terminating job callback thread')
            self.callback_thread.join()
        fastr.log.debug('DRMAA executor stopped!')

    def _queue_job(self, job):
        self.submit_queue.put(job, block=True)

    def _terminate_job(self, drmaa_job_id: int, retry=1):
        try:
            self.session.control(drmaa_job_id, drmaa.JobControlAction.TERMINATE)
        except (drmaa.InvalidJobException, drmaa.InternalException) as exception:
            new_status = self._get_job_state(drmaa_job_id)
            if new_status not in [drmaa.JobState.UNDETERMINED,
                                  drmaa.JobState.DONE,
                                  drmaa.JobState.FAILED]:
                if retry > 0:
                    self._terminate_job(drmaa_job_id, retry=retry - 1)
                    return
                else:
                    fastr.log.warning(f'Encountered a DRMAA exception: [{type(exception).__name__}]'
                                      f' {exception}')
        except Exception as exception:
            fastr.log.error(f'Encountered an exception during cancellation of job {drmaa_job_id}: '
                            f'[{type(exception).__name__}] {exception}')

    def _cancel_job(self, job, retry=True):
        with self.job_lookup_lock:
            drmaa_job_id = self.job_lookup_fastr_drmaa.pop(job.id, None)

        if drmaa_job_id is None:
            fastr.log.info('Job {} not found in DRMAA lookup, no longer exists?'.format(job.id))
            return

        fastr.log.debug('Cancelling job {}'.format(drmaa_job_id))
        self._terminate_job(drmaa_job_id)

        # FIXME: Should we call this here or will it called double? should we
        # just call it an make something to avoid double calls?
        self.job_finished(job, errors=[exceptions.FastrError('job cancelled by fastr DRMAA plugin').excerpt()])

    def _hold_job(self, job):
        drmaa_job_id = self.job_lookup_fastr_drmaa.get(job.id, None)
        if drmaa_job_id:
            try:
                self.session.control(drmaa_job_id, drmaa.JobControlAction.HOLD)
            except Exception as exception:
                fastr.log.error(f'Encountered an exception during setting job {drmaa_job_id} to hold: '
                                f'[{type(exception).__name__}] {exception}')
        else:
            fastr.log.error('Cannot hold job {}, cannot find the drmaa id!'.format(job.id))

    def _release_job(self, job):
        drmaa_job_id = self.job_lookup_fastr_drmaa.get(job.id, None)
        if drmaa_job_id:
            try:
                self.session.control(drmaa_job_id, drmaa.JobControlAction.RELEASE)
            except Exception as exception:
                fastr.log.error(f'Encountered an exception during releasing job {drmaa_job_id}: '
                                f'[{type(exception).__name__}] {exception}')
        else:
            fastr.log.error('Cannot release job {}, cannot find the drmaa id!'.format(job.id))

    def _job_finished(self, result):
        pass

    def create_native_spec(self, queue, walltime, memory, ncores, outputLog,
                           errorLog, hold_job, hold, work_dir):
        """
        Create the native spec for the DRMAA scheduler. Needs to be implemented
        in the subclasses

        :param str queue: the queue to submit to
        :param str walltime: walltime specified
        :param str memory: memory requested
        :param int ncores: number of cores requested
        :param str outputLog: the location of the stdout log
        :param str errorLog: the location of stderr log
        :param list hold_job: list of jobs to depend on
        :param bool hold: flag if job should be submitted in hold mode
        :return:
        """
        native_spec = []

        native_spec.append(self.spec_fields['WD'].format(workdir=work_dir))
        native_spec.append(self.spec_fields['QUEUE'].format(queue=queue))

        if walltime is not None:
            native_spec.append(self.spec_fields['WALLTIME'].format(walltime=walltime))

        if memory is not None:
            native_spec.append(self.spec_fields['MEMORY'].format(memory=memory))

        if ncores is not None and ncores > 1:
            native_spec.append(self.spec_fields['NCORES'].format(ncores=ncores))

        if outputLog is not None:
            native_spec.append(self.spec_fields['OUTPUTLOG'].format(outputlog=outputLog))

        if errorLog is not None:
            native_spec.append(self.spec_fields['ERRORLOG'].format(errorlog=errorLog))

        if hold_job is not None:
            if isinstance(hold_job, int):
                native_spec.append(self.spec_fields['DEPENDS'].format(hold_list=hold_job))
            elif isinstance(hold_job, list) or isinstance(hold_job, tuple):
                if len(hold_job) > 0:
                    jid_list = self.spec_fields['DEPENDS_SEP'].join([str(x) for x in hold_job])
                    native_spec.append(self.spec_fields['DEPENDS'].format(hold_list=jid_list))
            else:
                fastr.log.error('Incorrect hold_job type!')

        if hold:
            # Add a user hold to the job
            native_spec.append(self.spec_fields['HOLD'])

        return ' '.join(native_spec)

    # FIXME This needs to be more generic! This is for our SGE cluster only!
    def send_job(self, command, arguments, work_dir, queue=None, resources=None,
                 job_name=None, joinLogFiles=False,
                 outputLog=None, errorLog=None, hold_job=None, hold=False):
        # Create job template
        jt = self.session.createJobTemplate()
        jt.remoteCommand = command

        jt.args = arguments
        jt.joinFiles = joinLogFiles
        env = os.environ
        # Make sure environment modules do not annoy use with bash warnings
        # after the shellshock bug was fixed
        env.pop('BASH_FUNC_module()', None)
        env['PBS_O_INITDIR'] = work_dir
        jt.jobEnvironment = env

        if queue is None:
            queue = self.default_queue

        # Format resource if needed
        if resources.time:
            hours = resources.time // 3600
            minutes = (resources.time % 3600) // 60
            seconds = resources.time % 60

            walltime = '{}:{:02d}:{:02d}'.format(hours, minutes, seconds)
        else:
            walltime = None

        if resources.memory:
            memory = '{}M'.format(resources.memory)
        else:
            memory = None

        # Get native spec from subclass
        native_spec = self.create_native_spec(
            queue=queue,
            walltime=walltime,
            memory=memory,
            ncores=resources.cores,
            outputLog=outputLog,
            errorLog=errorLog,
            hold_job=hold_job,
            hold=hold,
            work_dir=work_dir,
        )

        fastr.log.debug('Setting native spec to: {}'.format(native_spec))
        jt.nativeSpecification = native_spec
        if job_name is None:
            job_name = command
            job_name = job_name.replace(' ', '_')
            job_name = job_name.replace('"', '')
            if len(job_name) > 32:
                job_name = job_name[0:32]

        jt.jobName = job_name

        # Send job to cluster
        job_id = self.session.runJob(jt)

        # Remove job template
        self.session.deleteJobTemplate(jt)

        return job_id

    def submit_jobs(self):
        while self.running:
            try:
                self.check_threads()

                # Max jobs is larger than zero (set) and less/equal than current jobs (full)
                if 0 < self.max_jobs <= self.n_current_jobs:
                    time.sleep(1)
                    continue

                job = self.submit_queue.get(block=True, timeout=2)

                # Get job command and write to file
                command = [sys.executable,
                           os.path.join(fastr.config.executionscript),
                           str(job.commandfile)]
                fastr.log.debug('Command to queue: {}'.format(command))

                # Make sure we do not submit after it stopped running
                if not self.running:
                    break

                fastr.log.debug('Queueing {} [{}] via DRMAA'.format(job.id, job.status))
                work_dir = os.path.abspath(os.path.dirname(job.commandfile))
                # Submit command to scheduler
                cl_job_id = self.send_job(command[0], command[1:], work_dir=work_dir,
                                          job_name='fastr_{}'.format(job.id),
                                          resources=job.resources,
                                          outputLog=job.stdoutfile,
                                          errorLog=job.stderrfile,
                                          hold_job=[self.job_lookup_fastr_drmaa[x] for x in job.hold_jobs if x in self.job_lookup_fastr_drmaa],
                                          hold=job.status == JobState.hold,
                                          )

                # Register job in the translation tables
                with self.job_lookup_lock:
                    self.job_lookup_fastr_drmaa[job.id] = cl_job_id
                self.submit_queue.task_done()
                fastr.log.info('Job {} queued via DRMAA as {}'.format(job.id, cl_job_id))

                # Set the queue lock to indicate there is content in the queue
            except queue.Empty:
                pass

        fastr.log.info('DRMAA submission thread ended!')

    def collect_jobs(self):
        while self.running:
            # Wait for the queue to contain
            try:
                self.check_threads()
                info = self.session.wait(drmaa.Session.JOB_IDS_SESSION_ANY, 2)
            except drmaa.ExitTimeoutException:
                pass  # Nothing there
            except drmaa.InvalidJobException:
                fastr.log.debug('No jobs left (session queue appears to be empty)')
                time.sleep(2)  # Sleep 2 seconds and try again
            except drmaa.NoActiveSessionException:
                self.running = False
                if not self.running:
                    fastr.log.debug('DRMAA session no longer active, quiting collector...')
                else:
                    fastr.log.critical('DRMAA session no longer active, but DRMAA executor not stopped properly! Quitting')
            except drmaa.errors.DrmaaException as exception:
                # Avoid the collector getting completely killed on another DRMAA exception
                fastr.log.warning('Encountered unexpected DRMAA exception: {}'.format(exception))
            except Exception as exception:
                if exceptions.get_message(exception).startswith('code 24:'):
                    # Avoid the collector getting completely killed this specific exception
                    # This is generally a job that got cancelled or something similar
                    fastr.log.warning('Encountered (probably harmless) DRMAA exception: {}'.format(exception))
                else:
                    fastr.log.error('Encountered unexpected exception: {}'.format(exception))
            else:
                self.finished_queue.put(info, block=True)

        fastr.log.info('DRMAA collect jobs thread ended!')

    def dispatch_callbacks(self):
        while self.running:
            try:
                self.check_threads()

                info = self.finished_queue.get(block=True, timeout=2)

                # Make sure we do not submit after it stopped running
                if not self.running:
                    break

                fastr.log.debug('Cluster DRMAA job {} finished'.format(info.jobId))

                # Create a copy of the job that finished and remove from the translation table
                errors = []
                with self.job_lookup_lock:
                    job_id = self.job_lookup_fastr_drmaa.inverse.pop(info.jobId, None)
                job = self.job_dict.get(job_id, None)

                if info.hasSignal:
                    errors.append(exceptions.FastrError('Job exited because of a signal, this might indicate it got killed because it attempted to use too much memory (or other resources)').excerpt())

                self.finished_queue.task_done()
                if job is not None:
                    # Send the result to the callback function
                    self.job_finished(job, errors=errors)
                else:
                    fastr.log.warning('Job {} no longer available (got cancelled/processed already?)'.format(info.jobId))
            except queue.Empty:
                pass

        fastr.log.info('DRMAA dispatch callback thread ended!')

    def _get_job_state(self, drmaa_job_id: int) -> "drmaa.JobState":
        try:
            return self.session.jobStatus(drmaa_job_id)
        except drmaa.errors.InvalidJobException:
            # Cannot find job, so status is undetermined
            return drmaa.JobState.UNDETERMINED
        except Exception as exception:
            fastr.log.error(f'Encountered an exception during getting state for job {drmaa_job_id}: '
                            f'[{type(exception).__name__}] {exception}')
            return drmaa.JobState.UNDETERMINED

    def regression_check(self):
        last_update = time.time()

        while self.running:
            self.check_threads()
            if time.time() - last_update < self.regression_check_interval:
                # If it is not time for an extra check, only check undetermined jobs
                # if required, otherwise just sleep
                if self.undetermined_jobs:
                    undetermined_jobs = list(self.undetermined_jobs.keys())

                    for drmaa_job_id in undetermined_jobs:
                        if not self.running:
                            break

                        self._job_regression_check(drmaa_job_id)
                else:
                    time.sleep(1)
                continue

            last_update = time.time()  # Reset timer

            jobs_to_check = list(self.job_lookup_fastr_drmaa.values())
            fastr.log.info(f'Running job regression check, {len(jobs_to_check)} job(s) to be checked')
            fastr.log.info(f'{self.submit_queue.qsize()} waiting to be submitted')
            fastr.log.info(f'{self.finished_queue.qsize()} waiting finished and waiting to be processed')
            fastr.log.info(f'{self.callback_queue.qsize()} waiting waiting for the final callback processing')

            for drmaa_job_id in jobs_to_check:
                if not self.running:
                    break

                self._job_regression_check(drmaa_job_id)

    def _job_regression_check(self, drmaa_job_id: int):
        status = self._get_job_state(drmaa_job_id)

        # Check for undetermined jobs
        if status == drmaa.JobState.UNDETERMINED:
            undetermined_count = self.undetermined_jobs.get(drmaa_job_id, 0) + 1

            # We do not consider the job lost yet, store count and continue
            if undetermined_count < self.num_undetermined_to_fail:
                self.undetermined_jobs[drmaa_job_id] = undetermined_count
                fastr.log.info(f'Job {drmaa_job_id} status could not determined {undetermined_count} times')
                return

        # No undetermined count that is of consequence anymore
        self.undetermined_jobs.pop(drmaa_job_id, 0)

        # Only start cleaning disappeared jobs
        if status not in [drmaa.JobState.DONE,
                          drmaa.JobState.FAILED,
                          drmaa.JobState.UNDETERMINED]:
            return

        fastr.log.info(f'Job {drmaa_job_id} is {status}, calling finished callback')

        # Job failed/done already!
        with self.job_lookup_lock:
            job_id = self.job_lookup_fastr_drmaa.inverse.pop(drmaa_job_id, None)
        job = self.job_dict.get(job_id, None)

        # We cannot get the fastr job related to this job, so we cannot continue
        if job is None:
            fastr.log.warning(f'Could not find connected fastr job for {drmaa_job_id}')
            return

        # Send the result to the callback function
        self.job_finished(
            job,
            errors=[exceptions.FastrError(
                "Job collected by the job status checking rather"
                " then getting a callback from DRMAA.").excerpt()])



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This package contains the base class and meta class for all serializable
objects in the Fastr system.
"""

import gzip
import json
import os
from abc import abstractmethod
from collections.abc import Mapping
import time
from pathlib import Path
from typing import Union, Optional, IO, Tuple

import yaml

from .. import exceptions, version
from ..helpers import xmltodict, jsonschemaparser, config, log


class ReadWriteHandler:
    HANDLERS = {}

    # Needs to be implemented in subclasses to register
    EXTENSIONS = ()

    def __init_subclass__(cls, **kwargs):
        for extension in cls.EXTENSIONS:
            cls.HANDLERS[extension] = cls

    @classmethod
    def get_handler(cls, method):
        return cls.HANDLERS[method]

    @staticmethod
    @abstractmethod
    def dump(doc: dict, stream: IO, **kwargs):
        """
        Dump doc into a stream

        :param stream: Stream to write to
        :param doc: Doc to write
        :param kwargs: Extra arguments for specific subclasses
        :return:
        """

    @staticmethod
    @abstractmethod
    def dumps(doc: dict, **kwargs) -> str:
        """
        Dump doc into a string

        :param doc: Doc to write
        :param kwargs: Extra arguments for specific subclasses
        :return: string representation of the doc
        """

    @classmethod
    def dumpf(cls, doc: dict, path: Union[str, os.PathLike], **kwargs):
        """
        Dump doc into a file

        :param doc: The data to write to the file
        :param path: Path of the file to write
        :param kwargs: Extra arguments for specific subclasses
        """
        if isinstance(path, os.PathLike):
            path = path.__fspath__()

        compress = path.endswith('.gz')

        if compress:
            open_func = gzip.open
        else:
            open_func = open

        with open_func(path, 'w') as fh_out:
            cls.dump(doc, fh_out, **kwargs)

            # Try to flush the file best as possible
            fh_out.flush()
            os.fsync(fh_out.fileno())

    @staticmethod
    @abstractmethod
    def load(stream: IO) -> dict:
        """
        Load object doc from a stream

        :param stream: stream to load from
        :return: the dict representation of the object read
        """

    @staticmethod
    @abstractmethod
    def loads(data: str) -> dict:
        """
        Load object doc from a string

        :param data: The string representation to load
        :return: The dict representation of the result
        """

    @classmethod
    def loadf(cls, path: Union[str, os.PathLike], retry_scheme: Tuple[int, ...]=(1, 3, 5)) -> dict:
        """
        Load object doc from a file

        :param path: Path of the file to load
        :return: The dict represention of the loaded object
        """
        if isinstance(path, Path):
            compress = path.suffix == '.gz'
        else:
            compress = path.endswith('.gz')

        if compress:
            open_func = gzip.open
        else:
            open_func = open

        attempt = 0

        # Loop for max retries
        while attempt <= len(retry_scheme):
            try:
                with open_func(path, 'rb') as fin:
                    data = cls.load(fin)
                break
            except Exception as exception:
                if attempt < len(retry_scheme):
                    delay = retry_scheme[0]
                    attempt += 1
                    log.debug('Retry {} after {} seconds of delay'.format(attempt, delay))
                    time.sleep(delay)
                elif isinstance(exception, FileNotFoundError):
                    raise exceptions.FastrFileNotFound(f'Could not open {path} for reading')
                else:
                    raise exception

        return data


class YamlHandler(ReadWriteHandler):
    """
    Read and write data to YAML files
    """
    EXTENSIONS = ('yaml', 'yml')

    @staticmethod
    def dump(data, stream, sort_keys=False, **kwargs):
        yaml.safe_dump(data, stream=stream, sort_keys=sort_keys, **kwargs)

    @staticmethod
    def dumps(data, sort_keys=False, **kwargs) -> str:
        return yaml.safe_dump(data, sort_keys=sort_keys, **kwargs)

    @staticmethod
    def load(data):
        return yaml.safe_load(data)

    @staticmethod
    def loads(data: str):
        return yaml.safe_load(data)


class JSONHandler(ReadWriteHandler):
    EXTENSIONS = ('json',)

    @staticmethod
    def dump(data, stream, **kwargs):
        json.dump(data, stream, **kwargs)

    @staticmethod
    def dumps(data, **kwargs) -> str:
        return json.dumps(data, **kwargs)

    @staticmethod
    def load(stream, **kwargs):
        return json.load(stream, **kwargs)

    @staticmethod
    def loads(data: str, **kwargs):
        return json.loads(data, **kwargs)


class XMLHandler(ReadWriteHandler):
    EXTENSIONS = ('xml',)

    @staticmethod
    def dump(data, stream):
        xmltodict.dump(data, stream)

    @staticmethod
    def dumps(data):
        return xmltodict.dumps(data)

    @staticmethod
    def load(stream):
        return xmltodict.load(stream)

    @staticmethod
    def loads(data: str):
        return xmltodict.loads(data)


class PassThroughHandler(ReadWriteHandler):
    """
    Passthrough handler for debugging of data just shows
    the string representation of the data
    """
    EXTENSIONS = ('dict',)

    @staticmethod
    def dump(stream, data):
        stream.write(repr(data))

    @staticmethod
    def dumps(data):
        return repr(data)

    @staticmethod
    def load(stream):
        raise NotImplementedError

    @staticmethod
    def loads(data: str):
        raise NotImplementedError


class Serializable:
    """
    Superclass for all classes that can be serialized.
    """
    SERIALIZERS = {}
    SERIALIZABLE_CLASSES = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        cls.SERIALIZABLE_CLASSES[cls.class_key()] = cls

    @classmethod
    def class_key(cls):
        return f'{cls.__module__}.{cls.__name__}'

    @classmethod
    def get_class(cls, key, default=None) -> 'Serializable':
        return cls.SERIALIZABLE_CLASSES.get(key, default)

    @classmethod
    def get_serializer(cls, filename=None):
        if filename is None:
            if hasattr(cls, '__dataschemafile__'):
                filename = cls.__dataschemafile__
            else:
                raise exceptions.FastrValueError('Cannot get serializer for class without dataschemafile specified!')
        if not os.path.exists(filename):
            filename = os.path.join(config.schemadir, filename)

        if filename not in cls.SERIALIZERS:
            try:
                cls.SERIALIZERS[filename] = jsonschemaparser.getblueprinter(filename)
            except IOError as err:
                log.warning('Serializable class {} cannot read schemafile {}: {}'.format(
                    cls.__name__, filename, err.strerror))

        return cls.SERIALIZERS[filename]

    @abstractmethod
    def __getstate__(self):
        raise exceptions.FastrNotImplementedError('For a Serializable the __getstate__ has to be set by the subjc')

    @classmethod
    def deserialize(cls, doc, network=None):
        """
        Classmethod that returns an object constructed based on the
        str/dict (or OrderedDict) representing the object

        :param dict doc: the state of the object to create
        :param network: the network the object will belong to
        :type network: :py:class:`Network <fastr.planning.network.Network>`
        :return: newly created object
        """
        return cls.createobj(doc, network)

    @classmethod
    def createobj(cls, state, _=None):
        """
        Create object function for generic objects

        :param cls: The class to create
        :param state: The state to use to create the Link
        :return: newly created Link
        """
        # If object is immutable with new args
        if hasattr(cls, '__getnewargs__'):
            if isinstance(state, Mapping):
                obj = cls.__new__(cls, **state)
            else:
                obj = cls.__new__(cls, *state)

        else:
            # Otherwise object can be created empty and set the state
            obj = cls.__new__(cls)

        if hasattr(obj, '__setstate__'):
            # We just check if __setstate__ existed, but pylint missed that
            # pylint: disable=no-member,maybe-no-member
            obj.__setstate__(state)
        elif hasattr(obj, '__dict__'):
            obj.__dict__.update(state)

        return obj

    def serialize(self):
        """
        Method that returns a dict/str/list structure with the data to rebuild
        the object.

        :returns: serialized representation of object
        """
        doc = self.__getstate__()
        return doc


def save(obj: Serializable, path: Union[str, os.PathLike], method: Optional[str] = None, annotate: bool = True):
    if isinstance(path, os.PathLike):
        path = path.__fspath__()

    if method is None:
        method = os.path.splitext(path)[1][1:]

    # Note the class used
    data = obj.serialize()
    if annotate:
        data['__class__'] = obj.class_key()
        data['__fastr_version__'] = version.full_version

    handler = ReadWriteHandler.get_handler(method)
    handler.dumpf(data, path)


def load(path: Union[str, os.PathLike], method: Optional[str] = None, cls=None):
    if isinstance(path, os.PathLike):
        path = path.__fspath__()

    if method is None:
        method = os.path.splitext(path)[1][1:]

    handler = ReadWriteHandler.get_handler(method)
    data = handler.loadf(path)

    if cls is None:
        class_spec = data.pop('__class__')

        if class_spec is None:
            raise exceptions.FastrValueError('Class is not added in the document, cannot find class to create')

        cls = Serializable.get_class(class_spec)

    fastr_version = data.pop('__fastr_version__', None)
    if fastr_version is None:
        log.debug(f'Loading objects without a fastr version')
    elif fastr_version != version.full_version:
        log.warning(f'Loading objects created with different version of fastr '
                    f'(current {version.full_version}, object saved with {fastr_version}')

    if hasattr(cls, '__dataschemafile__'):
        # We just check if __dataschemafile__ existed, but pylint missed that
        # pylint: disable=no-member
        data = cls.get_serializer().instantiate(data)

    # Instantiate object
    obj = cls.deserialize(data)

    # Store path
    obj.filename = path

    return obj



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime
import os
import json

import fastr
from fastr.helpers.classproperty import classproperty
from fastr.abc.baseplugin import PluginState
from fastr.plugins.reportingplugin import ReportingPlugin
from fastr.execution.job import JobState, Job
from fastr.exceptions import FastrOptionalModuleNotAvailableError


try:
    from elasticsearch import Elasticsearch
    elasticsearch_loaded = True
except ImportError:
    elasticsearch_loaded = False


class ElasticsearchReporter(ReportingPlugin):
    if not elasticsearch_loaded:
        _status = (PluginState.failed, 'Could not load elasticsearch module required for cluster communication')

    def __init__(self):
        super().__init__()

    @classmethod
    def test(cls):
        if not elasticsearch_loaded:
            raise FastrOptionalModuleNotAvailableError('Could not import the required elasticsearch for this plugin')

    def activate(self):
        """
        Activate the reporting plugin
        """
        super().activate()

        # Parse URI
        if fastr.config.elasticsearch_host == '':
            fastr.log.info("No valid elasticsearchsearch host given, elasticsearch Reporting will be disabled!")
            self.elasticsearch_uri = None
            return
        fastr.log.info("")
    
        self.elasticsearch_uri = fastr.config.elasticsearch_host
        self.elasticsearch_index = fastr.config.elasticsearch_index

        # Create Elastic search index 
        fastr.log.info("ES Logging to {} at index {}".format(self.elasticsearch_uri, self.elasticsearch_index))
        es = Elasticsearch([self.elasticsearch_uri])
        es.indices.create(index=self.elasticsearch_index, ignore=400)

    @classproperty
    def configuration_fields(cls):
        return {
            "elasticsearch_host": (str, "", "The elasticsearch host to report to"),
            "elasticsearch_index": (str, "fastr", "The elasticsearch index to store data in"),
            "elasticsearch_debug": (bool, False, "Setup elasticsearch debug mode to send stdout stderr on job succes"),
        }

    def elasticsearch_update_status(self, job):
        es = Elasticsearch([self.elasticsearch_uri])
        es.indices.create(index=self.elasticsearch_index, ignore=400)
        node = job.node
        job_data = {
            "timestamp": datetime.datetime.utcnow().isoformat(),
            "network_id" : node.parent.long_id,
            "network_version" : str(node.parent.network_version),
            "network_tmpurl" : node.parent.tmpurl,
            "run_id" : node.parent.id,
            "node" : str(node),
            "node_id" : node.id,
            "node_global_id" : node.global_id,
            "tool_name" : node.tool.ns_id,
            "tool_version" : str(node.tool.command['version']),
            "sample_index" : list(job.sample_index),
            "sample_id": list(job.sample_id),
            "errors": str(job.errors),
            "status": str(job.status),
        }
        if os.path.exists(job.extrainfofile):
            with open(job.extrainfofile) as extra_info_file:
                extra_info = json.load(extra_info_file)
                process = extra_info.get('process')
                job_data['process'] = process
        
        es.index(index=self.elasticsearch_index, doc_type='fastr-job', body=job_data)

    def job_updated(self, job: Job):
        if self.elasticsearch_uri:
            self.elasticsearch_update_status(job)

    # def run_started(self, run: NetworkRun):
    #     if self.elasticsearch_uri:
    #         self.api.elasticsearch_register_run(run)

    # def run_finished(self, run: NetworkRun):
    #     if self.elasticsearch_uri:
    #         self.api.elasticsearch_finish_run(run)

    # def log_record_emitted(self, record: LogRecord):
    #     if self.elasticsearch_uri:
    #         self.api.elasticsearch_log_line(record)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module to compare various fastr specific things such as a execution directory
or a reference directory.
"""

import os

import fastr
from fastr.execution.job import Job
from fastr.execution.networkanalyzer import DefaultNetworkAnalyzer
from fastr.execution.networkchunker import DefaultNetworkChunker
from fastr.helpers.iohelpers import load_gpickle


def compare_set(set1, set2, path, sub_compare_func, f_args=None, f_kwargs=None):
    """
    Compare two sets and dispatch each item to a sub comparison function

    :param Iterable set1: first set of items
    :param Iterable set2: second set of items
    :param str path: identifier of the data location
    :param sub_compare_func: function to apply to items
    :param f_args: args to pass to sub_compare_func
    :param f_kwargs: kwargs to pass to sub_compare_func
    :return: generator that iterates over the differences
    :rtype: generator
    """
    if not isinstance(set1, set):
        set1 = set(set1)

    if not isinstance(set2, set):
        set2 = set(set2)

    if f_args is None:
        f_args = ()

    if f_kwargs is None:
        f_kwargs = {}

    if set1 != set2:
        yield ("{} contains different keys, set 1 exclusively"
               " contains {} and set 2 exclusively contains {}").format(
            path,
            set1.difference(set2),
            set2.difference(set1)
        )

    for item in sorted(set1.intersection(set2)):
        for diff in sub_compare_func(item, *f_args, **f_kwargs):
            yield diff


def compare_execution_dir(path1, path2):
    # Compare network dumps
    network_file1 = os.path.join(path1, fastr.planning.network.Network.NETWORK_DUMP_FILE_NAME)
    network_file2 = os.path.join(path2, fastr.planning.network.Network.NETWORK_DUMP_FILE_NAME)

    fastr.log.info('Loading network 1: {}'.format(network_file1))
    network1 = fastr.planning.network.Network.loadf(network_file1)
    fastr.log.info('Loading network 2: {}'.format(network_file2))
    network2 = fastr.planning.network.Network.loadf(network_file2)

    fastr.log.debug('Network1 filename: {}'.format(network1.filename))
    fastr.log.debug('Network2 filename: {}'.format(network2.filename))

    if network1 != network2:
        yield "Networks dumps are not equal!"
    else:
        del network2

    # Get the order of the Nodes
    execution_order = []
    # Create execution objects
    chuncker = DefaultNetworkChunker()
    analyzer = DefaultNetworkAnalyzer()

    # Create a network chuncker to Chunk the Network in executable blocks
    chunks = chuncker.chunck_network(network1)

    for chunk in chunks:
        # Create a network analyzer to create the optimal execution order
        execution_order.extend(analyzer.analyze_network(None, chunk))

    fastr.log.debug('Execution order: "{}"'.format(execution_order))

    # Compare node outputs in execution order
    for node in execution_order:
        fastr.log.debug('Checking node {}'.format(node.id))
        # Get the sample present
        node_dir1 = os.path.join(path1, node.id)
        node_dir2 = os.path.join(path2, node.id)

        if isinstance(node, fastr.SourceNode):
            # Possible source nodes do not exist
            if not os.path.isdir(node_dir1):
                if not os.path.isdir(node_dir2):
                    # Non-existing in both
                    continue
                else:
                    yield("NodeRun '{}' does not have output for result 2")
                    continue
            elif not os.path.isdir(node_dir2):
                yield("NodeRun '{}' does not have output for result 1")
                continue

        for diff in compare_set(
                        os.listdir(node_dir1),
                        os.listdir(node_dir2),
                        node.id,
                        compare_job_dirs,
                        (node, node_dir1, node_dir2)):
            yield diff


def compare_job_dirs(sample, node, node_dir1, node_dir2):
    fastr.log.debug('Checking sample {}'.format(sample))
    result1 = os.path.join(node_dir1, sample, Job.RESULT_DUMP)
    result2 = os.path.join(node_dir2, sample, Job.RESULT_DUMP)

    result1_exists = os.path.exists(result1)
    result2_exists = os.path.exists(result2)

    if result1_exists and not result2_exists:
        yield "Ouput data for {} exists, but {} does not exist".format(
            result1,
            result2,
        )
    elif not result1_exists and result2_exists:
        yield "Ouput data for {} exists, but {} does not exist".format(
            result2,
            result1,
        )

    if result1_exists and result2_exists:
        job1 = load_gpickle(result1)
        job2 = load_gpickle(result2)

        for diff in compare_set(
                list(job1.output_data.keys()),
                list(job2.output_data.keys()),
                '{}/{}'.format(job1.node_id, job1.sample_id),
                compare_job_output_data,
                (job1, job2)):
            yield diff


def compare_job_output_data(output, job1, job2):
    # Compare output data
    data1 = job1.output_data[output]
    data2 = job2.output_data[output]

    if isinstance(data1, list):
        for diff in compare_value_list(data1, data2, '{}/{}'.format(job1.id, output)):
            yield diff
    else:
        for diff in compare_set(
                list(data1.keys()),
                list(data2.keys()),
                '{}/{}'.format(job1.node_id, job1.sample_id),
                compare_value_dict_item,
                (data1, data2, '{}/{}'.format(job1.id, output))):
            yield diff


def compare_value_dict_item(key, data1, data2, path):
    if key is not None:
        data1 = data1[key]
        data2 = data2[key]

    path = '{}/{}'.format(path, key)

    for diff in compare_value_list(data1, data2, path):
        yield diff


def compare_value_list(data1, data2, path, key=None):
    if key is not None:
        data1 = data1[key]
        data2 = data2[key]

    fastr.log.debug('Job1 data: {}'.format(data1))
    fastr.log.debug('Job2 data: {}'.format(data2))
    if len(data1) != len(data2):
        yield "Cardinality for {} differs ({} vs {})".format(
                path,
                len(data1),
                len(data2),
        )
    else:
        for index, (item1, item2) in enumerate(zip(data1, data2)):
            if item1 != item2:
                yield "Output data for path {}, index {} does not match ({} vs {})".format(
                        path,
                        index,
                        item1,
                        item2
                      )



# -*- coding: utf-8 -*-
#
# fastr documentation build configuration file, created by
# sphinx-quickstart on Fri Aug 12 17:01:23 2011.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

from collections import namedtuple
import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.dirname(__file__))

# Read the docs hacks
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    print('[conf.py] On Read the Docs')

from generate_modules import recurse_tree
from doc_clean import clean

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
        'sphinx.ext.autodoc',
        'sphinx_autodoc_typehints',
        'sphinx.ext.todo',
        'sphinx.ext.coverage',
        'sphinx.ext.viewcode',
        'sphinx.ext.inheritance_diagram',
        'sphinx.ext.intersphinx',
        'sphinx.ext.graphviz',
        'sphinxcontrib.httpdomain',
        'sphinxarg.ext'
]

# Intersphinx mapping to Pyhton documentation
intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}

# Graphviz output format to svg
graphviz_output_format = 'svg'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# Numbering references
numfig = True

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
# master_doc = 'modules'
master_doc = 'index'

# General information about the project.
project = 'FASTR'
copyright = '2011 -- 2019, Biomedical Imaging Group Rotterdam, Departments of ' \
            'Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#

try:
    import fastr.version
    release = fastr.version.version
except ImportError:
    with open(os.path.join(os.path.dirname(__file__),
                           '..', '..',
                           'version')) as version_file:
        release = version_file.read().strip()

# The short X.Y version.
version = '.'.join(release.split('.')[:2])
print('[conf.py] Found version: {}'.format(version))

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
if not on_rtd:
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
else:
    html_theme = 'default'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = 'static/images/logo.svg'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = 'static/images/favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['static/sources']

html_css_files = ['extra.css']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'FASTRdoc'


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
latex_element = {
    'papersize': 'a4',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howtos/manual]).
latex_documents = [
    ('index', 'FASTR.tex', 'FASTR Documentation',
     'Fastr contributors', 'manual'),
]

# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'bigr-fastr', 'FASTR Documentation',
     ['Fastr contributors'], 1)
]

# Some autogeneration of documentation (make autodoc index, generate plugin
# references etc.
if hasattr(sys, 'real_prefix'):
    print('[conf.py] Inside virtual env: {}'.format(sys.prefix))
else:
    print('[conf.py] Not inside a virtual env!')

try:
    import fastr
    print('[conf.py] Loaded fastr, will build reference of available plugins')
except ImportError:
    print('[conf.py] Cannot load fastr, will not build reference of plugins!')
    fastr = None

print('[conf.py] Current directory: {}'.format(os.path.abspath(os.curdir)))

print('[conf.py] Cleaning old files...')
clean()  # call doc_clean via code

if fastr is not None:
    print('[conf.py] Generating configuration reference')
    from generate_config import generate_config_doc
    generate_config_doc()

    print('[conf.py] Generating Plugin reference')
    from generate_plugins import generate_all
    generate_all()

    print('[conf.py] Generating Command-line reference')
    from generate_cmd import generate_command_reference
    generate_command_reference()

print('[conf.py] python generate_modules.py ..' + os.path.sep + ' -d .' + os.path.sep + ' -s rst -f')

# Call generate modules from code rather than via a subprocess
# Feed all options via opts namedtuple
options = {'destdir': './_autogen/',
           'dryrun': False,
           'force': True,
           'maxdepth': 4,
           'notoc': False,
           'suffix': 'rst'}

opts = namedtuple('opts', list(options.keys()))(*list(options.values()))
rootpath = '../'
excludes = [
    os.path.abspath(os.path.join(rootpath, 'doc')),
    os.path.abspath(os.path.join(rootpath, 'examples')),
    os.path.abspath(os.path.join(rootpath, 'resources')),
]
recurse_tree(rootpath, excludes, opts)

print('[conf.py] Done...')
print('[conf.py] Found files in curdir: {}'.format(os.listdir('.')))



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import weakref
from collections import OrderedDict

from sympy.core.symbol import Symbol

from ..abc.updateable import UpdateableMeta
from .. import exceptions
from ..core.dimension import HasDimensions, Dimension
from ..core.samples import SampleIndex, SampleItem, SamplePayload, SampleState
from ..helpers import log
from ..planning.inputoutput import BaseInput


class InputGroup(OrderedDict, HasDimensions, metaclass=UpdateableMeta):
    """
    A class representing a group of inputs. Input groups allow the
    """
    __updatetriggers__ = ['__init__',
                          '__setitem__',
                          '__delitem__',
                          'clear',
                          'pop',
                          'popitem',
                          'setdefault',
                          'update']

    def __init__(self, parent, id_=None):
        """
        Create a new InputGroup representation

        :param parent: the parent node
        :type parent: :py:class:`NodeRun <fastr.planning.node.NodeRun>`
        :param str id_: the id of the input group
        :raises FastrTypeError: if parent is not a NodeRun
        """
        super(InputGroup, self).__init__()
        self._parent = weakref.ref(parent)
        self.id = id_
        self._dimensions = ()
        self._primary = None
        self.__updating__ = True

    def __getitem__(self, key):
        if isinstance(key, SampleIndex):
            index = key

            # Determine which input sample to use for Input
            data = {}

            # Match dimensions if possible
            for id_, input_ in self.items():
                source_index = self.find_source_index(target_size=self.size,
                                                      target_dimnames=self.dimnames,
                                                      source_size=input_.size,
                                                      source_dimnames=input_.dimnames,
                                                      target_index=index)

                # Get data from Input
                if source_index is not None:
                    data[id_] = input_[source_index]
                else:
                    data[id_] = SampleItem(index, '__EMPTY__', [], set())

            # Aggregate data for input group
            sample_id = data[self.primary.id].id
            hold_jobs = set.union(*[val.jobs for val in data.values()])
            status = SampleState.combine(x.status for x in data.values())

            return SamplePayload(index, sample_id, data, hold_jobs, set(), status)
        else:
            return super(InputGroup, self).__getitem__(key)

    def __setitem__(self, key, value, dict_setitem=dict.__setitem__):
        """
        Assign an input to this input group.

        :param str key: id of the input
        :param value: the input to assign
        :type value: :py:class:`Input <fastr.planning.inputoutput.Input>`
        :raises FastrTypeError: if value of valid type
        """
        if not isinstance(value, BaseInput):
            message = 'Cannot assign a non-Input to an InputGroup'
            log.error(message)
            raise exceptions.FastrTypeError(message)

        if self.parent is not None and value.node is not self.parent:
            message = 'Input has a different parent NodeRun than the InputGroup, this is not a valid assignment!'
            log.error(message)
            raise exceptions.FastrParentMismatchError(message)

        super(InputGroup, self).__setitem__(key, value)

    @property
    def fullid(self):
        return '{}/input_groups/{}'.format(self.parent.fullid, self.id)

    @property
    def parent(self):
        """
        The parent node of this InputGroup
        """
        if self._parent is not None:
            return self._parent()
        else:
            return None

    @property
    def dimensions(self):
        """
        The dimensions of this InputGroup
        """
        return self._dimensions

    @property
    def primary(self):
        """
        The primary Input in this InputGroup. The primary Input is the Input
        that defines the size of this InputGroup. In case of ties it will be
        the first in the tool definition.
        """
        return self._primary

    @property
    def empty(self):
        """
        Bool indicating that this InputGroup is empty (has no data connected)
        """
        return self.size is None or len([x for x in self.size if x != 0]) == 0

    def find_source_index(self, target_size, target_dimnames, source_size, source_dimnames, target_index):
        # Determine which input sample to use for Input

        if source_size == target_size:
            source_index = target_index
        elif source_size == (1,):
            source_index = SampleIndex(0)
        elif all(x == 0 for x in source_size):
            source_index = None
        else:
            source_index = self.solve_broadcast(target_size=target_size,
                                                target_dimnames=target_dimnames,
                                                source_size=source_size,
                                                source_dimnames=source_dimnames,
                                                target_index=target_index,
                                                nodegroups=self.parent.parent.nodegroups)

        return source_index

    @staticmethod
    def _get_lookup(target_dimnames, nodegroups):
        lookup = {v: dimname for dimname in target_dimnames for value in nodegroups.values() if dimname in value for v in value}
        lookup.update({x: x for x in target_dimnames})

        return lookup

    @classmethod
    def solve_broadcast(cls, target_size, target_dimnames, source_size, source_dimnames, target_index, nodegroups=None):
        indexmap = dict(zip(target_dimnames, target_index))
        sizemap = dict(zip(target_dimnames, target_size))
        lookup = cls._get_lookup(target_dimnames, nodegroups)

        if all(x in lookup for x in source_dimnames):
            matched_dims = [lookup[x] for x in source_dimnames]
            source_index = SampleIndex(indexmap[x] for x in matched_dims)
            estimated_source_size = tuple(sizemap[x] for x in matched_dims)
        else:
            raise exceptions.FastrSizeMismatchError('Cannot broadcast, not all dimnames can be matched! (source dimnames {}, lookup {}'.format(source_dimnames, lookup))

        if source_size != estimated_source_size:
            raise exceptions.FastrSizeMismatchError('The estimated size after broadcast matching is incorrect! (estimated {}, actual {})'.format(estimated_source_size, source_size))

        return source_index

    @property
    def iterinputvalues(self):
        """
        Iterate over the item in this InputGroup

        :returns: iterator yielding :py:class:`SampleItems <fastr.core.sampeidlist.SampleItem>`
        """
        for sample_item in self.primary.items():
            yield self[sample_item.index]

    def __updatefunc__(self):
        """
        Update the InputGroup. Triggers when a change is made to the content
        of the InputGroup. Automatically recalculates the size, primary Input
        etc.
        """
        sizes = [x.size for x in self.values()]
        dimnames = [x.dimnames for x in self.values()]

        unique_sizes = set(sizes) - {(0,), (1,), (), None}

        if len(unique_sizes) > 1:
            if not any(all(not isinstance(x, Symbol) for x in size) for size in unique_sizes):
                # All entries in unique sizes are symbols, we cannot really
                # know what will be the size. Assume for now that the first
                # Input with symbolic input will be the primary
                max_dims = max(x.ndims for x in self.values())
                for input_ in self.values():
                    if input_.size in unique_sizes and input_.ndims == max_dims:
                        self._primary = input_
                        self._dimensions = input_.dimensions
                        break

            # Check if we can match via dimnames
            elif all(len(x) == len(set(x)) for x in dimnames):
                longest_dimname, longest_size = max(zip(dimnames, sizes), key=lambda x: len(x[1]))
                longest_dimensions = tuple(Dimension(name, size) for name, size in zip(longest_dimname, longest_size))

                if all(all(x in longest_dimname for x in y) for y in dimnames):
                    self._dimensions = longest_dimensions
                    self._primary = list(self.values())[sizes.index(self.size)]
                else:
                    nodegroups = self.parent.parent.nodegroups
                    lookup = {v: dimname for dimname in longest_dimname for value in nodegroups.values() if dimname in value for v in value}
                    lookup.update({x: x for x in longest_dimname})

                    if all(x in lookup for y in dimnames for x in y):
                        self._dimensions = longest_dimensions
                        self._primary = list(self.values())[sizes.index(self.size)]
                    else:
                        message = '[{}] Not all dimnames ({}) are present in the highest-dimensional input ({})'.format(
                            self.fullid,
                            [x for y in dimnames for x in y],
                            list(lookup.keys()),
                        )
                        log.warning(message)

                        self._dimensions = longest_dimensions
                        self._primary = list(self.values())[sizes.index(self.size)]
            else:
                message = 'One or more inputs have non-unique dimnames ({})'.format(dimnames)
                log.error(message)
                raise exceptions.FastrValueError(message)

        elif len(unique_sizes) == 1:
            self._primary = list(self.values())[sizes.index(unique_sizes.pop())]
            self._dimensions = self._primary.dimensions
        elif (1,) in sizes:
            self._primary = list(self.values())[sizes.index((1,))]
            self._dimensions = self._primary.dimensions
        else:
            self._primary = None
            self._dimensions = ()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict

import sympy

from .. import exceptions
from ..core.samples import SampleId, SampleItem, SampleIndex, SampleState
from .inputoutputrun import AdvancedFlowOutputRun, OutputRun
from .job import InlineJob, JobState
from .noderun import NodeRun
from ..helpers import log

__all__ = ['FlowNodeRun', 'AdvancedFlowNodeRun']


class FlowNodeRun(NodeRun):
    """
    A Flow NodeRun is a special subclass of Nodes in which the amount of samples
    can vary per Output. This allows non-default data flows.
    """
    _OutputType = OutputRun

    @property
    def blocking(self):
        """
        A FlowNodeRun is (for the moment) always considered blocking.

        :return: True
        """
        return True

    @property
    def outputsize(self):
        """
        Size of the outputs in this NodeRun
        """
        # Get sizes of all input groups
        output_size = []
        for input_group in self.input_groups.values():
            if input_group.size is not None:
                output_size.extend(input_group.size)
            else:
                return None

        output_size.append(sympy.symbols('N_{}'.format(self.id)))
        return tuple(output_size)

    @property
    def dimnames(self):
        """
        Names of the dimensions in the NodeRun output. These will be reflected
        in the SampleIdList of this NodeRun.
        """
        if self.nodegroup is not None:
            extra_dim = self.nodegroup
        else:
            extra_dim = self.id

        return super(FlowNodeRun, self).dimnames + (extra_dim,)

    def set_result(self, job, failed_annotation):
        """
        Incorporate result of a job into the FlowNodeRun.

        :param Type job: job of which the result to store
        """
        log.debug('Job output data: {}'.format(job.output_data))

        # Get the main sample index from the Job
        sample_index = job.sample_index

        for output in self.outputs.values():
            if output.id not in job.output_data:
                log.error('Could not find expected data for {} in {}!'.format(output.fullid, job.output_data))

            if failed_annotation:
                data = [(job.sample_id, (job.get_deferred(output.id, 0),))]
            else:
                data = job.output_data[output.id]

            log.debug('output_data = {}'.format(data))

            # Make sure dictionary is sorted, can also be list of items
            # which will be kept ordered
            if isinstance(data, dict):
                data = sorted(data.items())

            if not all(isinstance(x, (list, tuple)) and len(x) == 2 for x in data):
                raise exceptions.FastrValueError('The output data for a FlowNodeRun should be a dictionary or a list of items (length 2 per entry)')

            for sample_nr, (sample_id, sample_data) in enumerate(data):
                orig_sample_id = sample_id

                # Ensure we have a SampleId (cast if need be)
                if not isinstance(sample_id, SampleId):
                    # Make sure sample_id is built from a tuple of str
                    if isinstance(sample_id, str):
                        sample_id = (str(sample_id),)
                    else:
                        sample_id = tuple(str(x) for x in sample_id)

                    sample_id = SampleId(sample_id)

                    log.debug('Change sample_id from {} ({}) to {} ({})'.format(orig_sample_id,
                                                                                      type(orig_sample_id).__name__,
                                                                                      sample_id,
                                                                                      type(sample_id).__name__))

                if len(sample_id) != output.ndims:
                    sample_id = job.sample_id + sample_id
                    log.debug('Updated sample_id to {}'.format(sample_id))
                    if len(sample_id) != output.ndims:
                        raise exceptions.FastrValueError('Sample ID {} has the wrong dimensionality!'.format(sample_id))

                log.debug('Setting data for blocking node: {} sample: {}'.format(output.fullid, sample_id))

                output_values = tuple(job.get_deferred(output.id,
                                                       c,
                                                       orig_sample_id) for c, _ in enumerate(sample_data))

                log.debug('Setting collected for {} sample_id {} sample_index {!r} data: {}'.format(output.fullid,
                                                                                                          sample_id,
                                                                                                          sample_index + (sample_nr),
                                                                                                          output_values))

                # Save with sample_index and sample nr in the extra dimension
                output[sample_id, sample_index + (sample_nr)] = SampleItem(sample_index + (sample_nr),
                                                                           sample_id,
                                                                           OrderedDict({0: tuple(output_values)}),
                                                                           {job},
                                                                           failed_annotation)

                # Register the samples parent job
                self.jobs[sample_id] = job


class AdvancedFlowNodeRun(FlowNodeRun):
    _OutputType = AdvancedFlowOutputRun
    _JobType = InlineJob

    def _execute(self):
        """
        Execute the node and create the jobs that need to run

        :return: list of jobs to run
        :rtype: list of :py:class:`Jobs <fastr.execution.job.Job>`
        """
        self.update(False, False)

        # Make sure a NodeRun is valid
        if not self.valid:
            message = 'NodeRun {} is not valid'.format(self.fullid)
            log.error(message)
            log.error('Messages:\n{}'.format('\n'.join(self.messages)))
            raise exceptions.FastrNodeNotValidError(message)
        input_groups = self.input_groups

        data = {x.id: list(x.items()) for x in self.inputs.values()}
        target = self.tool.target

        job = self.create_job(SampleId('FLOW'),
                              SampleIndex(0),
                              job_data=data,
                              job_dependencies=None,
                              status=SampleState.VALID)

        with target:
            result = self.tool.interface.execute(target, data)

        job.flow_data = result.result_data

        output_data = {key: {str(v.id): v.data.sequence_part() for k, v in list(value.items())} for key, value in list(result.result_data.items())}
        job.output_data = output_data

        job.status = JobState.execution_done
        job.write()

        yield job

    def set_result(self, job, failed_annotation):
        for output, data in job.flow_data.items():
            log.debug('Advanced flow for output: {}'.format(output))
            for (sample_index, sample_id), value in data.items():
                log.debug('Advanced flow sample {!r} -> {}'.format(sample_index, list(value.data)))

                output_values = tuple(job.get_deferred(output,
                                                       c,
                                                       sample_id) for c, _ in enumerate(value.data))

                log.debug('Setting collected for {} sample_id {!r} sample_index {!r} data: {}'.format(output,
                                                                                                      sample_id,
                                                                                                      sample_index,
                                                                                                      output_values))

                # Save with sample_index and sample nr in the extra dimension
                self.outputs[output][sample_index] = SampleItem(value.index,
                                                                value.id,
                                                                OrderedDict({0: tuple(output_values)}),
                                                                {job},
                                                                failed_annotation)

        self.jobs['FLOW'] = job



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import re
import subprocess
import sys
from fastr.utils.cmd import add_parser_doc_link


def directory(path):
    """
    Make sure the path is a valid directory for argparse
    """
    path = os.path.abspath(os.path.expanduser(path))
    if not os.path.isdir(path):
        raise ValueError('Reference should be a valid directory')
    return path


def tool(value):
    """
    Make sure the value is a correct tool for argparse or reference directory
    """
    import fastr
    try:
        return fastr.tools[value]
    except fastr.exceptions.FastrError:
        return directory(value)


def check_tool(args):
    import fastr

    if isinstance(args.tool, str):
        tool_ref_file = os.path.join(args.tool, fastr.core.tool.Tool.TOOL_REFERENCE_FILE_NAME)
        if os.path.isfile(tool_ref_file):
            fastr.core.tool.Tool.test_tool(args.tool)
        else:
            fastr.log.warning('Could not find valid reference data in {}'.format(args.tool))
    else:
        try:
            fastr.log.info('Testing Tool {} {}'.format(args.tool.ns_id, args.tool.command_version))
            args.tool.test()
        except fastr.exceptions.FastrValueError as e:
            fastr.log.error('Tool is not valid: {}'.format(e))


def check_tools(args):
    import fastr

    for (_, _, _), tool in sorted(fastr.tools.items()):
        if tool.tests is not None and len(tool.tests) > 0:
            # Suppress logging during test
            old_level = fastr.config.loglevel
            fastr.config.loglevel = 50
            result = tool.test()
            fastr.config.loglevel = old_level

            if len(result) > 0:
                fastr.log.info('{}/{}: FAILED'.format(tool.ns_id, tool.command_version))
            else:
                fastr.log.info('{}/{}: PASSED'.format(tool.ns_id, tool.command_version))


def check_network(args):
    import fastr

    # Test if it is a Network or Tool reference directory
    network_file = os.path.join(args.reference, fastr.Network.NETWORK_DUMP_FILE_NAME)

    if os.path.isfile(network_file):
        fastr.log.info('Found Network reference data, testing network')
        fastr.Network.test(args.reference)
    else:
        fastr.log.warning('Could not find valid network reference data in {}'.format(args.reference))


def check_networks(args):
    import fastr

    # Try to test all subdirectories (use them as reference dir)
    subdirs = sorted(x for x in os.listdir(args.reference) if os.path.isdir(os.path.join(args.reference, x)))
    results = []

    for subdir in sorted(subdirs):
        network_dir = os.path.join(args.reference, subdir)

        print('Testing {}... '.format(network_dir)),
        sys.stdout.flush()

        # Make sure the directory is a valid test directory
        if not os.path.exists(os.path.join(network_dir, fastr.Network.NETWORK_DUMP_FILE_NAME)):
            print('SKIPPED')
            continue

        # Create a result entry
        result = {
            'name': subdir,
            'reference': network_dir,
        }

        results.append(result)

        # Run the test in a subprocess (to isolate tests and
        # avoid resource hogging between tests)
        command = ['fastr', 'test', 'network', network_dir]
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()

        # Save results
        result['stdout'] = stdout
        result['stderr'] = stderr
        result['return_code'] = process.returncode

        # Check if result is valid
        passed = len(stderr) == process.returncode == 0
        passed = passed and re.search(r'^\[INFO\] *network:\d\d\d\d >> Run and reference were equal! Test passed!$',
                                      stdout,
                                      re.MULTILINE) is not None

        result['passed'] = passed
        if len(stderr) == 0:
            print('PASSED' if passed else 'FAILED')
        else:
            print('ERROR')

    if args.result:
        from fastr.helpers.iohelpers import save_json
        save_json(args.result, results)

    if all(x['passed'] for x in results):
        print('ALL TEST PASSED')
    else:
        print('ENCOUNTERED FAILURES')

    return results


def get_parser():
    parser = argparse.ArgumentParser(description="Run a tests for a fastr resource.")
    subparsers = parser.add_subparsers(help='There are different resources that can be tested')

    tool_parser = subparsers.add_parser('tool', help='Test a single tool')
    tool_parser.add_argument('tool', metavar='TOOL', type=tool,
                             help='Tool to test or directory with tool reference data')
    tool_parser.set_defaults(func=check_tool)

    tools_parser = subparsers.add_parser('tools', help='Test all tools known to fastr')
    tools_parser.set_defaults(func=check_tools)

    network_parser = subparsers.add_parser('network', help='Test a single network')
    network_parser.add_argument('reference', metavar='NETWORK', type=directory,
                                help='The reference data to test the Network')
    network_parser.set_defaults(func=check_network)

    networks_parser = subparsers.add_parser('networks', help='Test all network references inside subdirectories')
    networks_parser.add_argument('reference', metavar='REFERENCE', type=directory,
                                 help='path of the directory containing subdirectories with reference data')
    networks_parser.add_argument('--result', metavar='RESULT.json', type=str, required=False,
                                 help='Write the results of the test to a JSON file')
    networks_parser.set_defaults(func=check_networks)
    return parser


def main():
    """
    Run the tests of a tool to verify the proper function
    """
    # No arguments were parsed yet, parse them now
    parser = add_parser_doc_link(get_parser(), __file__)
    args, unknown_args = parser.parse_known_args()

    # Ship off subparser to correct function
    args.func(args)


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from fastr.core import vfs_plugin
from fastr.core.dimension import Dimension
from fastr.core.samples import SampleItem
from .inputoutputrun import MacroOutputRun
from .job import SinkJob, Job, JobState
from .networkrun import NetworkRun
from .noderun import NodeRun
from fastr.helpers import log

__all__ = ['MacroNodeRun']


class MacroNodeRun(NodeRun):
    """
    MacroNodeRun encapsulates an entire network in a single node.
    """
    _OutputType = MacroOutputRun

    def __init__(self, node, parent):
        """
        :param network: network to create macronode for
        :type network: fastr.planning.network.Network
        """
        super(MacroNodeRun, self).__init__(node=node, parent=parent)

        self._network_run = NetworkRun(node.network)
        self.network_run.parent = self

    @property
    def network_run(self):
        return self._network_run

    def __getstate__(self):
        """
        Retrieve the state of the MacroNodeRun

        :return: the state of the object
        :rtype dict:
        """
        state = super(MacroNodeRun, self).__getstate__()
        state['network_run'] = self._network_run.__getstate__()
        return state

    def __setstate__(self, state):
        self._network_run = NetworkRun.deserialize(state.pop('network_run'))
        super(MacroNodeRun, self).__setstate__(state)

    def get_output_info(self, output):
        source_dimensions = {}

        for input_ in self.inputs.values():
            source_node = self.network_run.sourcelist[input_.id]
            source_dimension_name = source_node.dimnames[0]
            source_dimensions[source_dimension_name] = input_.dimensions

        # Translate back result index and id
        sink = self.network_run.sinklist[output.id]
        new_dimensions = []

        for dimname, size in zip(sink.dimnames, sink.outputsize):
            # If they were translated, replace them back
            if dimname in source_dimensions:
                dimensions_part = source_dimensions[dimname]

                new_dimensions.extend(x.copy() for x in dimensions_part)
            else:
                new_dimensions = Dimension(dimname, size)

        return tuple(new_dimensions)

    def execute(self):
        # Should we check validity of the node and inside network again?
        # Prepare the output of the Node
        source_data = {}
        sink_data = {}
        network_run = self.network_run

        # The data required to map the linearized samples back to original format
        source_dimensions = {}

        # Set environment for network
        network_run.executing = True
        network_run.tmpdir = os.path.join(self.parent.tmpdir, self.id)
        network_run.tmpurl = vfs_plugin.path_to_url(network_run.tmpdir)
        network_run.timestamp = self.parent.timestamp

        for input_ in self.inputs.values():
            source_node = network_run.sourcelist[input_.id]

            # Register dimension name for source
            sample_mapping = {}

            if source_node.dimnames[0] not in source_dimensions:
                source_dimensions[source_node.dimnames[0]] = sample_mapping

            for index, sample_item in enumerate(input_.items()):
                sample_mapping[index] = sample_item.index, sample_item.id

                new_sample = sample_item.replace(
                    index=index,
                    id='+'.join(sample_item.id),
                )

                source_node.output[new_sample] = new_sample

                dummy_job = Job(source_node,
                                sample_item.id,
                                sample_item.index,
                                [],
                                [])
                dummy_job.status = JobState.finished

            # Data is put straight in the output, add empty dummy
            source_data[input_.id] = []

        for output in self.outputs.values():
            # Set datatype for sink to match requested datatype on output
            self.network_run.sinklist[output.id].datatype = output.resulting_datatype

            # Set sink data to ignore
            sink_data[output.id] = 'null://ignore/sinks'

        # Set the data for the network
        network_run.set_data(source_data, sink_data)

        # Start generating jobs
        for job_list in network_run.generate_jobs():
            # Stop if execution ended
            if not self.parent.executing:
                network_run.executing = self.parent.executing
                return

            yield (x for x in job_list if not isinstance(x, SinkJob))

        for sink in network_run.sinklist.values():
            sink.update()  # Make sure the sink is updated
            log.debug('Getting results from sink {} with dimnames: {}'.format(sink.global_id, sink.dimnames))
            output = self.outputs[sink.id]

            # Copy sink data to node
            for sample_item in sink.input.items():

                new_id = []
                new_index = []

                # Translate back result index and id
                for dimname, index_part, id_part in zip(sink.dimnames, sample_item.index, sample_item.id):
                    # If they were translated, replace them back
                    if dimname in source_dimensions:
                        index_part, id_part = source_dimensions[dimname][index_part]
                        new_index.extend(index_part)
                        new_id.extend(id_part)
                    else:
                        new_index.append(index_part)
                        new_id.append(id_part)

                new_sample = sample_item.replace(
                    index=new_index,
                    id=new_id
                )

                log.debug("Setting {} in {}".format(new_sample, output.fullid))
                output[new_sample] = new_sample

                dummy_job = Job(sink,
                                sample_item.id,
                                sample_item.index,
                                [],
                                [])
                dummy_job.status = JobState.finished

        # Only now the NodeRun is drained of all jobs
        self.drained = True



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gzip
import json
# Try to import the faster cPickle if available and fall back when needed
import pickle as pickle
import os
import shutil
import time

from . import log
from .filesynchelper import FileSyncHelper, filesynchelper_enabled
from ..resources import tools


def load_gpickle(path, retry_scheme=None):
    if filesynchelper_enabled():
        file_sync_helper = FileSyncHelper()
        if file_sync_helper.has_file_promise(path):
            # If we use filesynchelper we don't care about the
            # pickle on the filesystem. Use the filesynchelper
            # cache instead.
            data = file_sync_helper.load(path)
            try:
                result = pickle.loads(data)
                return result
            except AttributeError as exception:
                # This is probably due to an Enum that is created on the fly
                # when a Tool is loaded
                log.warning('Cannot find attribute during unpickling,'
                             ' this is probably an extension type that'
                             'was created with an older fastr version, '
                             'attempting to populate resources and try '
                             'again. Original exception: {}'.format(exception))
                tools.populate()
                return pickle.loads(data)

    if retry_scheme is None:
        retry_scheme = (1, 3, 5)

    log.debug('Attempting to load {} with retry scheme {}'.format(path, retry_scheme))

    # Start with a non-delayed attempt
    retry_scheme = (0,) + tuple(retry_scheme)

    # Get the jobfile if given a job
    for attempt, delay in enumerate(retry_scheme, start=1):
        try:
            # Wait before trying to load the file
            log.debug('Attempt {} after {} seconds of delay'.format(attempt, delay))
            time.sleep(delay)

            try:
                with gzip.open(path, 'rb') as fh_in:
                    data = pickle.load(fh_in)
            except AttributeError as exception:
                # This is probably due to an Enum that is created on the fly
                # when a Tool is loaded
                log.warning('Cannot find attribute during unpickling,'
                            ' this is probably an extension type that'
                            'was created with an older fastr version, '
                            'attempting to populate resources and try '
                            'again. Original exception: {}'.format(exception))
                tools.populate()
                with gzip.open(path, 'rb') as fh_in:
                    data = pickle.load(fh_in)

            return data
        except Exception as exception:
            # Last retry failed
            if attempt == len(retry_scheme):
                raise exception


def save_gpickle(path, data):
    with gzip.open(path, 'wb') as fh_out:
        pickle.dump(data, fh_out)

        # Make sure the data gets flushed
        fh_out.flush()
        os.fsync(fh_out.fileno())

    if filesynchelper_enabled():
        file_sync_helper = FileSyncHelper()
        # if we use filesynchelper store the pickle
        # in there as well
        file_sync_helper.store(path, pickle.dumps(data))


def save_json(path, data, indent=2):
    with open(path, 'w') as fh_out:
        # If possible use fastr serialization
        if hasattr(data, 'dump'):
            data.dump(fh_out, method='json', indent=indent)
        else:
            json.dump(data, fh_out, indent=indent)

        # Make sure the data gets flushed
        fh_out.flush()
        os.fsync(fh_out.fileno())

    if filesynchelper_enabled():
        FileSyncHelper().store(path, json.dumps(data))


def load_json(path):
    if filesynchelper_enabled():
        file_sync_helper = FileSyncHelper()
        if file_sync_helper.has_file_promise(path):
            return json.loads(file_sync_helper.load(path))

    with open(path, 'r') as in_fh:
        return json.load(in_fh)


def link_or_copy(source: str, destination: str):
    try:
        os.symlink(source, destination)
    except OSError:
        shutil.copy2(source, destination)




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import fastr
import os


def teardown():
    pass


def test_guesstype_url_noextension():
    pass


def test_guesstype_url_extension():
    mhd_file_mhd = os.path.join(fastr.config.mounts['example_data'], 'images', 'mrwhite.mhd')
    mhd_type = fastr.types.guess_type(fastr.vfs.path_to_url(mhd_file_mhd))
    assert mhd_type is fastr.types['MetaImageFile']


def test_guesstype_path_noextension():
    pass


def test_guesstype_path_extension():
    pass


def test_guesstype_options_issubclass_typegroup():
    pass


def test_guesstype_options_value_exception():
    pass


# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

FASTR is a top level package which includes all parts required to create
networks and edit networks.
"""

__all__ = [
    'create_network',
    'create_network_copy',
    'config',
    'datatypes',
    'exceptions',
    'networks',
    'plugins',
    'tools',
    'types',
    'vfs',
    'MISSING',
]
# In the top level module we want to add some variables which are constants
# but use a non-constant name (not caps)
# pylint: disable=invalid-name


from colorama import init
init()

# Get version info
from . import version, exceptions
__version__ = version.version

#: Configuration and logging of the fastr system
from .helpers import config
from .helpers import log

# Load datatypes base before the API, as we need the MISSING here first
from . import datatypes

MISSING = datatypes.Missing()

# Load API
from .api import create_network, create_network_copy

# Import data types and create a lazy loading module for that
from .helpers.lazy_module import LazyModule

types = datatypes.types
datatypes = LazyModule("datatypes", parent=datatypes, plugin_manager=types)

# Import plugins and create a lazy loading module for that
from . import plugins
plugin_manager = plugins.plugins
plugins = LazyModule("plugins", parent=plugins, plugin_manager=plugins.plugins)
plugin_manager.set_module(plugins)

# Load resources for tools and networks
from .resources import tools, networks

# The following loads all ioplugins from the resources folder and registers the built-in vfs with the plugin list
from .core import vfs_plugin as vfs

log.debug('Finished with the FASTR environment set up')


# Warn if this is not a neatly installed package from the stable branch
if config.warn_develop and (version.not_master_branch or version.from_git):
    log.warning('Not running in a production installation (branch "{}" from {})'.format(
        version.git_branch,
        'source code' if version.from_git else 'installed package'
    ))
